/*
   Projet de simulation d'inondation.
   
   Version C++ utilisant QT et OpenGL.
   
   10 Avril 2004

   (C) Eric Beaudry & Universit de Sherbrooke 2004.
*/


#include "SimInondation.h"
#include <math.h>
#include <iostream>


Mesh::Mesh(int w, int h) : width(w), height(h) {
  sommets = new Point[w*h];
  vecteurs = new Vecteur[w*h];
  
  int w2 = w/2;
  int h2 = h/2;
  for(int x=0;x<w;x++)
    for(int y=0;y<h;y++){
       float xp = x-w2;
       float yp = y-h2;
       float zp = 0;
       sommets[x+y*width] = Point(xp, yp, zp);
    }
    
  autoVN();
}

Mesh::~Mesh(){
  delete[] sommets;
  delete[] vecteurs;
}

void Mesh::autoVN() {
  Vecteur vt;
  for(int x=0;x<width;x++)
    for(int y=0;y<height;y++){
       Point s = getSommet(x, y);
       Vecteur vn;
       
       if(x>0 && y>0){
            vt = (s-getSommet(x-1, y)) ^ (s-getSommet(x, y-1));
            vt.normalise();
            vn += vt;
          }
        
	if(x>0 && y<height-1){
            vt = (s-getSommet(x-1, y)) ^ (getSommet(x, y+1)-s);
            vt.normalise();
            vn += vt;
        }
        
        if(x<width-1 && y>0){
            vt = (getSommet(x+1, y)-s) ^ (s-getSommet(x,  y-1));
            vt.normalise();
            vn += vt;
        }

        if(x<width-1 && y<height-1){
            vt = (getSommet(x+1, y)-s) ^ (getSommet(x, y+1)-s);
            vt.normalise();
            vn += vt;
        }

        //vn.normalise();
	vecteurs[x + y * width] = vn;
    }
}


SimInondWidget::SimInondWidget(Terrain* t, QWidget* parent, const char* name) 
 : QGLWidget(parent, name)
{
    meshEau = 0;
    meshSol = 0;
    setTerrain(t);
    
    angleRotation = 0;
    angleSol = 30;
    distanceCamera = 100;
    refX = refY = 0;
    
    shiftDown = false;
    setFocusPolicy(QWidget::ClickFocus);
    
    //startTimer( 25 );
}

SimInondWidget::~SimInondWidget(){
    delete meshSol;
    delete meshEau;
}

void SimInondWidget::setTerrain(Terrain* t){
    delete meshSol;
    delete meshEau;
    
    terrain = t;
    
    meshSol = new Mesh(terrain->getWidth(), terrain->getHeight());
    meshEau = new Mesh(terrain->getWidth(), terrain->getHeight());
    
    double d1, d2;
    terrain->getMinMax(d1, d2);
    minTerrain = d1;
    maxTerrain = d2;
    
    float res = terrain->getResolution();
    
    int w2 = terrain->getWidth()/2;
    int h2 = terrain->getHeight()/2;
    for(int x=0;x<terrain->getWidth();x++)
      for(int y=0;y<terrain->getHeight();y++){
         meshSol->getSommet(x, y) = Point((x-w2)*res, (y-h2)*res, terrain->sol(x, y)-minTerrain);
	 meshEau->getSommet(x, y) = Point((x-w2)*res, (y-h2)*res, terrain->total(x, y)-minTerrain);
      }
    meshSol->autoVN();
    meshEau->autoVN();
}

void SimInondWidget::updateMeshEau(){
    for(int x=0;x<terrain->getWidth();x++)
      for(int y=0;y<terrain->getHeight();y++){
	 meshEau->getSommet(x, y).z = terrain->total(x, y) - minTerrain;
      }
    meshEau->autoVN();
}

void SimInondWidget::rotation(float angle){
   angleRotation += angle;
}

void SimInondWidget::initializeGL()
{
    static GLfloat pos[4] = {1, 0, -6.0, 1.0 };
    glLightfv( GL_LIGHT0, GL_POSITION, pos );
    glEnable( GL_CULL_FACE );
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glEnable( GL_NORMALIZE );
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
    //glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_ALPHA_TEST);
    glShadeModel(GL_SMOOTH);
}

void SimInondWidget::resizeGL(int width, int height)
{
    GLfloat w = (float) width / (float) height;
    GLfloat h = 1.0;

    glViewport( 0, 0, width, height );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum( -w, w, -h, h, 4.0, 1000.0 );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void SimInondWidget::timerEvent(QTimerEvent*)
{
    updateGL();
}


void SimInondWidget::paintGL()
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPushMatrix();
      glTranslatef(0, 0, - distanceCamera);
      glRotatef(angleSol-90, 1, 0, 0);
      glRotatef(angleRotation, 0,0,1);
      glTranslatef(refX, refY, 0);
      
    glColor3f(0.8, 0.8, 0.8);
    
    float scaleVert = 1 / (maxTerrain-minTerrain);

    //glEnable(GL_TEXTURE_2D);
    for(int x=0;x<meshSol->getWidth()-1;x++){
      glBegin(GL_QUAD_STRIP);
      //glBegin(GL_TRIANGLE_STRIP);
      for(int y=0;y<meshSol->getHeight();y++){
         const Point& p1 = meshSol->getSommet(x, y);
	 const Vecteur& v1 = meshSol->getVN(x, y);
	 const Point& p2 = meshSol->getSommet(x+1, y);
	 const Vecteur& v2 = meshSol->getVN(x+1, y);
	 
	 glNormal3f(v1.x, v1.y, v1.z);
	 glColor3f(0.8, 0.6 + p1.z * scaleVert, 0.8);
	 //glTexCoord2i(x, y);
	 glVertex3f(p1.x, p1.y, p1.z);
	 
	 glNormal3f(v2.x, v2.y, v2.z);
	 glColor3f(0.8, 0.6 + p2.z * scaleVert, 0.8);
	 //glTexCoord2i(x+1, y);
	 glVertex3f(p2.x, p2.y, p2.z);
      }
      glEnd();
    }
    //glDisable(GL_TEXTURE_2D);

    glEnable(GL_BLEND);    
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    for(int x=0;x<meshEau->getWidth()-1;x++){
      glBegin(GL_QUAD_STRIP);
      //glBegin(GL_TRIANGLE_STRIP);
      for(int y=0;y<meshEau->getHeight();y++){
         const Point& p1 = meshEau->getSommet(x, y);
	 const Vecteur& v1 = meshEau->getVN(x, y);
	 const Point& p2 = meshEau->getSommet(x+1, y);
	 const Vecteur& v2 = meshEau->getVN(x+1, y);
	 
	 glColor4f(0, 0, 1.0, terrain->eau(x, y));
	 glNormal3f(v1.x, v1.y, v1.z);
	 glVertex3f(p1.x, p1.y, p1.z);
	 
	 glColor4f(0, 0, 1.0, terrain->eau(x+1, y));
	 glNormal3f(v2.x, v2.y, v2.z);
	 glVertex3f(p2.x, p2.y, p2.z);
      }
      glEnd();
    }
    glDisable(GL_BLEND);
    
    /*
    glEnable(GL_BLEND);    
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(0.0, 0.0, 1.0, 0.5);
    glBegin(GL_POLYGON);
      glNormal3f(0, 0, 1);
      glVertex3f(20, -20, 2);
      glVertex3f(20, 20, 2);
      glVertex3f(-20, 20, 2);
      glVertex3f(-20, -20, 2);
    glEnd();
    glDisable(GL_BLEND);
    */
    glPopMatrix();
}

void SimInondWidget::mouseMoveEvent(QMouseEvent* e)
{
   int dx = lastPoint.x() - e->x();
   int dy = lastPoint.y() - e->y();
   if(e->state() &  Qt::LeftButton){
    if(shiftDown){
      distanceCamera += dy;
      if(distanceCamera<10) distanceCamera = 10;
      else if(distanceCamera>1000) distanceCamera = 1000;
    }else{
      angleSol += dy * 0.25;
      if(angleSol < 0) angleSol = 0;
      else if(angleSol > 90) angleSol = 90;
      angleRotation += -dx * 0.25;
    }
    updateGL();
   }
   
   if(e->state() & Qt::RightButton){
      double a = 3.141592654 * angleRotation / 180.0;
      refX += (-cos(a) * dx + sin(a) * dy) * 0.25;
      refY += (cos(a) * dy + sin(a) * dx) * 0.25;
      updateGL();
   }
   lastPoint = e->pos();
}

void SimInondWidget::mousePressEvent(QMouseEvent* e){
   lastPoint = e->pos();
}

void SimInondWidget::keyPressEvent(QKeyEvent* e){
   if(e->key() == Qt::Key_Shift)
      shiftDown = true;
}

void SimInondWidget::keyReleaseEvent(QKeyEvent* e){
   if(e->key() == Qt::Key_Shift)
      shiftDown = false;
}

void SimInondWidget::setTexture(const QImage& image){
   texture = QGLWidget::convertToGLFormat(image);
   glTexImage2D( GL_TEXTURE_2D, 0, 3, texture.width(), texture.height(), 0,
                  GL_RGBA, GL_UNSIGNED_BYTE, texture.bits() );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
}

